﻿#region License
/* Copyright 2011 James F. Bellinger <http://www.zer7.com/>

   This software is provided 'as-is', without any express or implied
   warranty. In no event will the authors be held liable for any damages
   arising from the use of this software.
 
   Permission is granted to anyone to use this software for any purpose,
   including commercial applications, and to alter it and redistribute it
   freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you must
   not claim that you wrote the original software. If you use this software
   in a product, an acknowledgement in the product documentation would be
   appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and must
   not be misrepresented as being the original software.

   3. This notice may not be removed or altered from any source
   distribution. 
*/
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;

namespace IllusoryStudios.Wpf.LostControls
{
    public class FitStackPanel : Panel
    {
        public static readonly DependencyProperty LayoutProperty =
            DependencyProperty.RegisterAttached("Layout", typeof(FitLayout),
            typeof(FitStackPanel), new FrameworkPropertyMetadata(null,
                FrameworkPropertyMetadataOptions.AffectsParentArrange|
                FrameworkPropertyMetadataOptions.AffectsParentMeasure|
                FrameworkPropertyMetadataOptions.AffectsArrange|
                FrameworkPropertyMetadataOptions.AffectsMeasure));

        IEnumerable<KeyValuePair<UIElement, Rect>> ComputeSizes(Size finalSize)
        {
            // Measure the desired sizes for the various types.
            double autoSpan = 0, pixelSpan = 0;
            double shrinkableSpan = 0, shrinkableDesiredSpan = 0;
            double starSpan = 0, starCount = 0;

            foreach (UIElement child in InternalChildren)
            {
                FitLayout layout = GetLayout(child);
                double minHeight = layout == null ? 0 : layout.Minimum;

                if (layout == null || layout.LayoutType == FitLayoutType.Auto)
                {
                    autoSpan += Math.Max(child.DesiredSize.Height, minHeight);
                }
                else if (layout.LayoutType == FitLayoutType.Pixel)
                {
                    pixelSpan += Math.Max(layout.Value, minHeight);
                }
                else if (layout.LayoutType == FitLayoutType.Shrinkable)
                {
                    shrinkableDesiredSpan += Math.Max(child.DesiredSize.Height, minHeight);
                    shrinkableSpan += minHeight;
                }
                else if (layout.LayoutType == FitLayoutType.Star)
                {
                    starCount += layout.Value;
                    starSpan += minHeight;
                }
            }

            // Now, how much will we actually have for Shrinkable and Star?
            double availableSpan = finalSize.Height;
            double reservedSpan = autoSpan + pixelSpan + shrinkableSpan + starSpan;
            double freeSpan = Math.Max(0, availableSpan - reservedSpan);

            // Free span goes first to Shrinkable, and the rest goes to Star.
            double shrinkableDesiredFreeSpan = shrinkableDesiredSpan - shrinkableSpan;
            double shrinkableFinalFreeSpan = Math.Min(shrinkableDesiredFreeSpan, freeSpan);
            double starFinalFreeSpan = freeSpan - shrinkableFinalFreeSpan;

            // Time to arrange!
            double position = 0;

            foreach (UIElement child in InternalChildren)
            {
                FitLayout layout = GetLayout(child);
                double minHeight = layout == null ? 0 : layout.Minimum;
                double finalSpan = 0;

                if (layout == null || layout.LayoutType == FitLayoutType.Auto)
                {
                    finalSpan = Math.Max(child.DesiredSize.Height, minHeight);
                }
                else if (layout.LayoutType == FitLayoutType.Pixel)
                {
                    finalSpan = Math.Max(layout.Value, minHeight);
                }
                else if (layout.LayoutType == FitLayoutType.Shrinkable)
                {
                    double extraNeeded = Math.Max(child.DesiredSize.Height, minHeight) - minHeight;
                    finalSpan = minHeight + (shrinkableDesiredFreeSpan > 0
                        ? extraNeeded * shrinkableFinalFreeSpan / shrinkableDesiredFreeSpan : 0);
                }
                else if (layout.LayoutType == FitLayoutType.Star)
                {
                    finalSpan = minHeight + starFinalFreeSpan * layout.Value / starCount;
                }

                var finalRect = new Rect(0, position, finalSize.Width, finalSpan);
                yield return new KeyValuePair<UIElement, Rect>(child, finalRect);
                position += finalSpan;
            }
        }

        protected override Size ArrangeOverride(Size finalSize)
        {
            foreach (var kvp in ComputeSizes(finalSize).ToList())
            {
                kvp.Key.Arrange(kvp.Value);
            }

            return finalSize;
        }

        protected override Size MeasureOverride(Size availableSize)
        {
            double span = 0, altSpan = 0;

            foreach (UIElement child in InternalChildren)
            {
                child.Measure(availableSize);
                altSpan = Math.Max(altSpan, child.DesiredSize.Width);
                span += child.DesiredSize.Height;
            }

            Size finalSize = new Size
                (Math.Min(availableSize.Width, altSpan),
                 Math.Min(availableSize.Height, span));
            foreach (var kvp in ComputeSizes(finalSize).ToList())
            { // Surprising that I need this. From my reading of the docs, I shouldn't. Ah well.
                kvp.Key.Measure(kvp.Value.Size);
            }

            return finalSize;
        }

        public static FitLayout GetLayout(UIElement element)
        {
            return (FitLayout)element.GetValue(LayoutProperty);
        }

        public static void SetLayout(UIElement element, FitLayout value)
        {
            element.SetValue(LayoutProperty, value);
        }
    }
}
