﻿#region License
/* Copyright 2010 James F. Bellinger <http://www.zer7.com/>

   This software is provided 'as-is', without any express or implied
   warranty. In no event will the authors be held liable for any damages
   arising from the use of this software.
 
   Permission is granted to anyone to use this software for any purpose,
   including commercial applications, and to alter it and redistribute it
   freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you must
   not claim that you wrote the original software. If you use this software
   in a product, an acknowledgement in the product documentation would be
   appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and must
   not be misrepresented as being the original software.

   3. This notice may not be removed or altered from any source
   distribution. 
*/
#endregion

using System;
using System.Collections.Specialized;
using System.Diagnostics;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Media;

namespace IllusoryStudios.Wpf.LostControls
{
    public class ThumbnailPanel : VirtualizingPanel, IScrollInfo
    {
        public static readonly DependencyProperty GridSizeProperty = DependencyProperty.Register
            ("GridWidth", typeof(Size), typeof(ThumbnailPanel),
            new PropertyMetadata(HandleGridSizeChanged));

        bool _canHScroll, _canVScroll;
        Size _extent, _viewport; Vector _offset;
        ScrollViewer _owner;

        int _gridColumns, _gridRows, _gridRowsPerPage;
        Size _gridPageSize; int _eltStart, _eltEnd;

        public ThumbnailPanel()
        {
            //var children = InternalChildren; // WPF bug: must use getter first for generator to init
            ComputeExtents();
        }

        public void SetHorizontalOffset(double offset)
        {
            offset = Math.Max(0, Math.Min(offset, _extent.Width - _viewport.Width));
            if (_offset.X != offset) { _offset.X = offset; InvalidateArrange(); UpdateElements(false); ScrollUpdate(); }
        }

        public double HorizontalOffset
        {
            get { return _offset.X; }
        }

        public void SetVerticalOffset(double offset)
        {
            offset = Math.Max(0, Math.Min(offset, _extent.Height - _viewport.Height));
            if (_offset.Y != offset) { _offset.Y = offset; InvalidateArrange(); UpdateElements(false); ScrollUpdate(); }
        }

        public double VerticalOffset
        {
            get { return _offset.Y; }
        }

        public double ViewportHeight
        {
            get { return _viewport.Height; }
        }

        public double ViewportWidth
        {
            get { return _viewport.Width; }
        }

        IItemContainerGenerator GetGenerator()
        {
            var dontCare = InternalChildren; // WPF bug: ItemContainerGenerator can be null if we don't use this getter first
            return ItemContainerGenerator;
        }

        int GetElementCount()
        {
            var ic = ItemsControl.GetItemsOwner(this);
            int count = ic != null ? ic.Items.Count : 0;
            return count;
        }

        void GetElementRange(out int start, out int end)
        {
            int count = GetElementCount();
            int startRow = Math.Max(0, -1 + (int)Math.Floor(_offset.Y / GridSize.Height));
            int endRow = Math.Max(0, 1 + (int)Math.Ceiling((_offset.Y + ViewportHeight) / GridSize.Height));
            start = Math.Min(count, startRow * _gridColumns);
            end = Math.Min(count, start + (endRow - startRow + 1) * _gridColumns);
        }

        Size ComputeExtents() // based on the current viewport
        {
            int count = GetElementCount();
            _gridColumns = Math.Max(1, Math.Min(1000, (int)Math.Floor(_viewport.Width / GridSize.Width)));
            _gridRows = (count + _gridColumns - 1) / _gridColumns;
            _gridRowsPerPage = Math.Max(1, Math.Min(1000, (int)Math.Floor(_viewport.Height / GridSize.Height)));
            _gridPageSize = new Size(_gridColumns * GridSize.Width, _gridRowsPerPage * GridSize.Height);
            return new Size(GridSize.Width * _gridColumns, GridSize.Height * _gridRows);
        }

        void GetGridRowCol(int index, out int row, out int col)
        {
            row = index / _gridColumns; col = index % _gridColumns;
        }

        Point GetGridPosition(int index)
        {
            int row, col; GetGridRowCol(index, out row, out col);
            return new Point(GridSize.Width * col, GridSize.Height * row);
        }

        protected override Size ArrangeOverride(Size finalSize)
        {
            Update(finalSize);
            var g = GetGenerator();

            for (int i = 0; i < InternalChildren.Count; i++)
            {
                var index = g.IndexFromGeneratorPosition(new GeneratorPosition(i, 0)); Debug.Assert(index >= 0);
                InternalChildren[i].Arrange(new Rect(GetGridPosition(index) - _offset, GridSize));
            }

            return finalSize;
        }

        protected override Size MeasureOverride(Size availableSize)
        {
            Update(availableSize); UpdateElements(false);

            foreach (UIElement child in InternalChildren)
            {
                child.Measure(GridSize);
            }

            return new Size
                (double.IsPositiveInfinity(availableSize.Width) ? 65535 : availableSize.Width,
                double.IsPositiveInfinity(availableSize.Height) ? 65535 : availableSize.Height);
        }

        protected override void OnItemsChanged(object sender, ItemsChangedEventArgs args)
        {
            base.OnItemsChanged(sender, args);

            switch (args.Action)
            {
                case NotifyCollectionChangedAction.Remove:
                    RemoveInternalChildRange(args.Position.Index, args.ItemUICount);
                    break;

                case NotifyCollectionChangedAction.Reset:
                    RemoveInternalChildRange(0, InternalChildren.Count);
                    break;
            }

            UpdateElements(true);
        }

        void Update(Size viewport)
        {
            if (_viewport != viewport) { _viewport = viewport; ScrollUpdate(); }
            UpdateExtents();
        }

        void UpdateExtents()
        {
            Size extent = ComputeExtents();
            if (_extent != extent) { _extent = extent; ScrollUpdate(); }
        }

        void UpdateElements(bool force)
        {
            int start, end; GetElementRange(out start, out end);
            if (start == _eltStart && end == _eltEnd && !force) { return; }
            _eltStart = start; _eltEnd = end;

            var g = GetGenerator();
            {
                var pos = g.GeneratorPositionFromIndex(start);
                using (g.StartAt(pos, GeneratorDirection.Forward, true))
                {
                    int childIndex = pos.Index + (pos.Offset > 0 ? 1 : 0);
                    for (int i = start; i < end; i++, childIndex++)
                    {
                        bool isNew;
                        var child = (UIElement)g.GenerateNext(out isNew);
                        if (!isNew) { continue; }

                        InsertInternalChild(childIndex, child);
                        g.PrepareItemContainer(child);
                    }
                }
            }

            for (int i = InternalChildren.Count - 1; i >= 0; i--)
            {
                var pos = new GeneratorPosition(i, 0);
                var index = g.IndexFromGeneratorPosition(pos);
                int extraItems = _gridColumns * _gridRowsPerPage;
                if (index >= start - extraItems && index < end + extraItems) { continue; }

                g.Remove(pos, 1); RemoveInternalChildRange(i, 1);
            }

            // In case filtering has removed many items, let's scroll ourselves...
            UpdateExtents();
            SetHorizontalOffset(HorizontalOffset);
            SetVerticalOffset(VerticalOffset);
        }

        void ScrollUpdate()
        {
            if (_owner != null) { _owner.InvalidateScrollInfo(); }
        }

        public Size GridSize
        {
            get { return (Size)GetValue(GridSizeProperty); }
            set { SetValue(GridSizeProperty, value); }
        }

        static void HandleGridSizeChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            ((ThumbnailPanel)d).InvalidateMeasure();
        }

        #region IScrollInfo
        bool IScrollInfo.CanHorizontallyScroll
        {
            get { return _canHScroll; }
            set { _canHScroll = value; }
        }

        bool IScrollInfo.CanVerticallyScroll
        {
            get { return _canVScroll; }
            set { _canVScroll = value; }
        }

        double IScrollInfo.ExtentHeight
        {
            get { return _extent.Height; }
        }

        double IScrollInfo.ExtentWidth
        {
            get { return _extent.Width; }
        }

        void IScrollInfo.LineDown()
        {
            SetVerticalOffset(VerticalOffset + GridSize.Height);
        }

        void IScrollInfo.LineLeft()
        {
            SetHorizontalOffset(HorizontalOffset - GridSize.Width);
        }

        void IScrollInfo.LineRight()
        {
            SetHorizontalOffset(HorizontalOffset + GridSize.Width);
        }

        void IScrollInfo.LineUp()
        {
            SetVerticalOffset(VerticalOffset - GridSize.Height);
        }

        Rect IScrollInfo.MakeVisible(Visual visual, Rect rectangle)
        {
            int index = InternalChildren.IndexOf((UIElement)visual);
            if (index == -1) { return rectangle; }

            var g = GetGenerator();
            index = g.IndexFromGeneratorPosition(new GeneratorPosition(index, 0));
            if (index == -1) { return rectangle; }

            var r = new Rect(GetGridPosition(index), GridSize);
            if (r.Top < VerticalOffset) { SetVerticalOffset(r.Top); }
            else if (r.Bottom > VerticalOffset + ViewportHeight) { SetVerticalOffset(r.Bottom - ViewportHeight); }

            if (r.Left < HorizontalOffset) { SetHorizontalOffset(r.Left); }
            else if (r.Right > HorizontalOffset + ViewportWidth) { SetHorizontalOffset(r.Right - ViewportWidth); }

            return rectangle;
        }

        void IScrollInfo.MouseWheelDown()
        {
            SetVerticalOffset(VerticalOffset + GridSize.Height);
        }

        void IScrollInfo.MouseWheelLeft()
        {
            SetHorizontalOffset(HorizontalOffset - GridSize.Width);
        }

        void IScrollInfo.MouseWheelRight()
        {
            SetHorizontalOffset(HorizontalOffset + GridSize.Width);
        }

        void IScrollInfo.MouseWheelUp()
        {
            SetVerticalOffset(VerticalOffset - GridSize.Height);
        }

        void IScrollInfo.PageDown()
        {
            SetVerticalOffset(VerticalOffset + _gridPageSize.Height);
        }

        void IScrollInfo.PageLeft()
        {
            SetHorizontalOffset(HorizontalOffset - _gridPageSize.Width);
        }

        void IScrollInfo.PageRight()
        {
            SetHorizontalOffset(HorizontalOffset + _gridPageSize.Width);
        }

        void IScrollInfo.PageUp()
        {
            SetVerticalOffset(VerticalOffset - _gridPageSize.Height);
        }

        ScrollViewer IScrollInfo.ScrollOwner
        {
            get { return _owner; }
            set { _owner = value; }
        }
        #endregion
    }
}
