﻿#region License
/* Copyright 2010 James F. Bellinger <http://www.zer7.com/>

   This software is provided 'as-is', without any express or implied
   warranty. In no event will the authors be held liable for any damages
   arising from the use of this software.
 
   Permission is granted to anyone to use this software for any purpose,
   including commercial applications, and to alter it and redistribute it
   freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you must
   not claim that you wrote the original software. If you use this software
   in a product, an acknowledgement in the product documentation would be
   appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and must
   not be misrepresented as being the original software.

   3. This notice may not be removed or altered from any source
   distribution. 
*/
#endregion

using System;
using System.ComponentModel;
using System.Linq;
using System.Reflection;
using Microsoft.Win32;

namespace IllusoryStudios.Wpf.LostControls.Win32
{
    public class StartupSettings : INotifyPropertyChanged
    {
        const string RegistryPath = @"Software\Microsoft\Windows\CurrentVersion\Run";

        public event PropertyChangedEventHandler PropertyChanged;

        string _desiredPath;

        static T CheckNull<T>(string name, T value)
        {
            if (value == null) { throw new ArgumentNullException("name"); }
            return value;
        }

        public StartupSettings(string startupKey)
        {
            CheckNull("startupKey", startupKey);
            if (startupKey.Length == 0) { throw new ArgumentException("Cannot be zero-length.", "startupKey"); }

            StartupKey = startupKey; DesiredPath = CurrentPath;
        }

        public void SetDesiredAssembly(Assembly assembly, params string[] args)
        {
            CheckNull("assembly", assembly);
            if (assembly.Location.Length == 0) { throw new ArgumentException("Assembly is not on disk.", "assembly"); }
            SetDesiredCommandLine(new[] { assembly.Location }.Concat(args).ToArray());
        }

        public void SetDesiredCommandLine(params string[] args)
        {
            CheckNull("args", args);
            DesiredPath = string.Join(" ", args.Select(x =>
                FileAssociationHelper.QuoteStringElement(x)).ToArray());
        }

        protected void OnPropertyChanged(string propertyName)
        {
            CheckNull("propertyName", propertyName);

            var pc = PropertyChanged;
            if (pc != null) { pc(this, new PropertyChangedEventArgs(propertyName)); }
        }

        public string CurrentPath
        {
            get
            {
                try
                {
                    var key = Registry.CurrentUser.OpenSubKey(RegistryPath);
                    return key.GetValue(StartupKey, null) as string ?? "";
                }
                catch { return ""; }
            }

            set
            {
                try
                {
                    var key = Registry.CurrentUser.OpenSubKey(RegistryPath, true);
                    if (string.IsNullOrEmpty(value)) { key.DeleteValue(StartupKey); }
                    else { key.SetValue(StartupKey, value); }
                }
                catch { }

                OnPropertyChanged("CurrentPath");
                OnPropertyChanged("StartsWithSystem");
            }
        }

        public string DesiredPath
        {
            get { return _desiredPath; }
            set
            {
                _desiredPath = value ?? "";
                OnPropertyChanged("DesiredPath");
                OnPropertyChanged("StartsWithSystem");
            }
        }

        public string StartupKey
        {
            get;
            private set;
        }

        public bool StartsWithSystem
        {
            get { return CurrentPath == DesiredPath; }
            set { CurrentPath = value ? DesiredPath : ""; }
        }
    }
}
